#!/usr/bin/python
# MIT License
#
# Copyright (c) 2017 Javier Romero
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# -*- coding: utf-8 -*-


#
# IMPORTS
#

import clingo
import asprin.src.solver.metasp.metasp_programs as metasp_programs
import asprin.src.solver.metasp.reify as reify
import sys
import re


#
# DEFINES
#

STATE_G = 0
STATE_C = 1
HELP = """\
usage: gc.py [--binary] [--check-to-sat] [number] [options] [guess_files] -C [check_files]
"""
ANSWER = """\
Answer {}:
{}"""
HOLDS = "holds"
BINDING_PYTHON = """
##true(atom(A)) :-     holds(X), ##output(holds(X),A).
##fail(atom(A)) :- not holds(X), ##output(holds(X),A).
:- not ##bot.
"""
BINDING_BINARY = """
##true(atom(A)) :-     holds(X), ##output(holds(X),B), ##literal_tuple(B,A).
##fail(atom(A)) :- not holds(X), ##output(holds(X),B), ##literal_tuple(B,A).
:- not ##bot.
"""


#
# METHODS
#

class Observer:

    def __init__(self):
        self.rules         = []
        self.weight_rules  = []
        self.output_atoms  = []
        self.output_terms  = []

    def rule(self, choice, head, body):
        self.rules.append((choice, head, body))

    def weight_rule(self, choice, head, lower_bound, body):
        self.weight_rules.append((choice, head, lower_bound, body))

    def output_atom(self, symbol, atom):
        self.output_atoms.append((symbol, atom))

    def output_term(self, symbol, condition):
        self.output_terms.append((symbol, condition))


def parse_args():
    options, guess, check = [], [], []
    state = STATE_G
    binary = False
    toSat = False
    append_next = False
    for i in sys.argv[1:]:
        if append_next:
            options.append(i)
            append_next=False
            continue
        if i == "--help":
            print(HELP)
            sys.exit(0)
        if i == "--binary":
            binary = True
        elif i == "--check-to-sat":
            toSat = True
        elif i == "-C":
            state = STATE_C
        elif i.startswith("-c"):
            options.append(i)
            append_next=True
        elif i.startswith("-") or i.isdigit():
            options.append(i)
        elif state == STATE_G:
            guess.append(i)
        else:
            check.append(i)
#    print (options)
    return options, binary, toSat, guess, check


def observe(program):
    observer = Observer()
    control = clingo.Control()
    control.register_observer(observer, replace=True)
    control.add("base", [], program)
    control.ground([("base", [])])
    return observer


def get_prefix(control):
    prefix = "_"
    for name, _, _ in control.symbolic_atoms.signatures:
        if name.startswith(prefix):
            prefix = re.sub(r'(_+).*', r'\1', name) + "_"
    return prefix

def gc_controls(guess, check, maxcontrol=2, reify_mode="observer", stats=False):
    assert reify_mode in ["observer", "binary", "sat"]

    options = ["0", "--heuristic=Domain", "--dom-mod=false,show",
            "--enum-mode=domRec",
            "-c", "maxcontrol={}".format(maxcontrol)]
    if stats:
        options.append("--stats")

    # start
    control = clingo.Control(options)

    # load guess and ground
    control.add("base", [], guess)
    control.ground([("base", [])])

    # get choices
    choices = "{ " + "; ".join(
        str(atom.symbol) for atom in
        control.symbolic_atoms.by_signature(HOLDS, 1)
    ) + " }."
    choices = "" if choices == "{  }." else choices

    # create check
    check += choices + "\n#show holds/1.\n"

    # fix prefix
    prefix = get_prefix(control)

    # reify check
    if reify_mode == "binary":
        check_reified  = reify.reify_from_string(check, prefix)
        check_reified += BINDING_BINARY.replace("##", prefix)
    elif reify_mode == "sat":
        check_reified  = reify.reify_from_string_through_sat(check, prefix)
        check_reified += BINDING_BINARY.replace("##", prefix)
    else:
        observer = observe(check)
        check_reified  = reify.reify_from_observer(observer, prefix)
        check_reified += BINDING_PYTHON.replace("##", prefix)
    # add meta encoding
    check_reified += metasp_programs.metaD_program.replace("##", prefix)

    # add to control and ground
    control.add("check", [], check_reified)
    control.ground([("check", [])])

    # solve
    models = 0
    with control.solve(yield_=True) as handle:
        for m in handle:
            yield m.symbols(shown=True)

    if stats:
        print("Time: ",control.statistics["accu"]["times"]["cpu"],"s", sep="")


def run():

    # start
    options, binary, toSat, guess_files, check_files = parse_args()
    control = clingo.Control(options)

    # load guess and ground
    for _file in guess_files:
        control.load(_file)
    control.ground([("base", [])])

    # get choices
    choices = "{ " + "; ".join(
        str(atom.symbol) for atom in
        control.symbolic_atoms.by_signature(HOLDS, 1)
    ) + " }."
    choices = "" if choices == "{  }." else choices

    # create check
    check = choices + "\n#show holds/1.\n"
    for _file in check_files:
        with open(_file, 'r') as f:
            check += f.read() + "\n"

    # fix prefix
    prefix = get_prefix(control)

    # reify check
    if binary:
        check_reified  = reify.reify_from_string(check, prefix)
        check_reified += BINDING_BINARY.replace("##", prefix)
    elif toSat:
        check_reified  = reify.reify_from_string_through_sat(check, prefix)
        check_reified += BINDING_BINARY.replace("##", prefix)
    else:
        observer = observe(check)
        check_reified  = reify.reify_from_observer(observer, prefix)
        check_reified += BINDING_PYTHON.replace("##", prefix)
    # add meta encoding
    check_reified += metasp_programs.metaD_program.replace("##", prefix)

    # add to control and ground
    control.add("check", [], check_reified)
    control.ground([("check", [])])

    # solve
    models = 0
    with control.solve(yield_=True) as handle:
        for m in handle:
            models += 1
            model = " ".join([str(x) for x in m.symbols(shown=True)])
            print(ANSWER.format(models, model))
        print(handle.get())
    if "--stats" in options:
        print("Time: ",control.statistics["accu"]["times"]["cpu"],"s", sep="")

if __name__ == "__main__":
    run()



"""
TODO
"""

from colomoto.minibn import BooleanNetwork

from algorecell_types import *

from .gc import gc_controls

def v(n):
    return "v{}".format(n)
def to_n(v):
    return v[1:]

def asp_of_condition(state):
    def from_cond(ns):
        n, s = ns
        pre = "not " if not s else ""
        return "{}{}".format(pre, v(n))
    return ", ".join(map(from_cond, sorted(state.items())))

def load(bn, inputs={}):
    """
    TODO
    """
    return ActoNet(bn, inputs)

class ActoNet(object):
    """
    TODO
    """
    def __init__(self, bn, inputs={}):
        bn = BooleanNetwork.auto_cast(bn)
        # TODO: handle constants
        assert not bn.constants(), \
            "Boolean network with constant nodes are not yet supported"
        self.bn = bn
        self.inputs = []
        self.outputs = []
        self.property = None
        self.set_input_condition(inputs)

    def _assert_defined(self, nodes):
        for n in nodes:
            assert n in self.bn, "unknown variable '{}'".format(n)

    def set_outputs(self, outputs):
        self._assert_defined(outputs)
        self.outputs = outputs

    def set_input_condition(self, state):
        self._assert_defined(state)
        self.input_conditions = state
        self.inputs = set(self.input_conditions.keys())

    def set_property(self, state):
        self._assert_defined(state)
        self.property = state

    def reprogramming_fixpoints(self, *spec, ignore=[],
            maxsize=5, **kwspec):
        """
        TODO
        """

        self.set_outputs(ignore)
        self.set_property(dict(*spec, **kwspec))

        strategies = ReprogrammingStrategies()
        global_alias = False
        inputs = self.input_conditions
        for c in self.controls(maxcontrol=maxsize):
            if type(c) is tuple:
                init = strategies.autoalias("input{}", c[0])
                p = PermanentPerturbation(c[1])
                s = FromCondition(init, PermanentPerturbation(pert))
            else:
                p = PermanentPerturbation(c)
                global_alias = True
                s = FromCondition("input", p) if inputs else FromAny(p)
            strategies.add(s)
        if global_alias and inputs:
            strategies.register_alias("input", inputs)
        return strategies


    def asp_fixpoint(self):
        clauses = []

        def from_literal(l):
            if l[0] in self.inputs:
                pre = "not " if not l[1] else ""
                return "{}holds({})".format(pre, v(l[0]))
            pre = "not " if l[1] else ""
            return "{}not {}".format(pre, v(l[0]))
        def from_clause(c):
            return ", ".join(map(from_literal, c))

        for n, cs in sorted(self.bn.as_dnf().items()):
            if n in self.inputs:
                continue
            is_output = n in self.outputs
            if not is_output:
                clauses.append("{0} :- holds(set({0}, 1)).".format(v(n)))
            for c in cs:
                pre = "not holds(set({}, _)), ".format(v(n)) if not is_output else ""
                clauses.append("{} :- {}{}.".format(v(n), pre, from_clause(c)))

        return clauses

    def asp_guess(self):
        clauses = [
            "value(0;1).",
            "variable({}).".format("; ".join(map(v, sorted(self.bn.keys())))),
            "output({}).".format("; ".join(map(v, sorted(self.outputs)))),
            "controlled(X) :- variable(X), not output(X).", # XXX controllable variables
            "control(X,T) :- controlled(X), value(T).",
            "#const maxcontrol=2.",
            "{set(X,V) : control(X,V)} maxcontrol.",
            ":- set(X,0), set(X,1).",
            "holds(set(X,V)) :- set(X,V).",
            "#show holds/1.",
            "#show nholds/1.",
            "#project holds/1.",
        ]
        #
        # inputs
        #
        clauses.append("{{{}}}.".format("; ".join(map(v, self.inputs))))
        for n in self.inputs:
            clauses.append("holds({0}) :- {0}.".format(v(n)))
            clauses.append("nholds({0}) :- not holds({0}).".format(v(n)))
        # TODO: ??
        if self.input_conditions:
            clauses.append("inputselectionnecessary :- {}.".format(asp_of_condition(self.input_conditions)))
            clauses.append(":- not inputselectionnecessary.")
        return clauses

    def asp_check(self):
        clauses = []
        if self.property:
            clauses.append(":- {}.".format(asp_of_condition(self.property)))
        return clauses

    def controls(self, universal=True, existential=True, maxcontrol=5):
        assert universal or existential, "At least universal or existential should be true"
        if universal:
            has_fixpoint = self.asp_fixpoint()
            guess = self.asp_guess()
            if existential:
                guess += has_fixpoint
            check = self.asp_check() + has_fixpoint
            guess = "\n".join(guess)
            check = "\n".join(check)
            for control in gc_controls(guess, check, maxcontrol=maxcontrol):
                c = {}
                m = {}
                for f in control:
                    assert f.name in ["holds", "nholds"]
                    g = f.arguments[0]
                    if g.name == "set" and len(g.arguments) == 2:
                        assert f.name == "holds"
                        m[to_n(g.arguments[0].name)] = g.arguments[1].number
                    else:
                        c[to_n(g.name)] = 1 if f.name == "holds" else 0
                if len(self.input_conditions) == len(self.inputs):
                    yield m
                else:
                    yield (c, m)
        else:
            raise NotImplementedError




from setuptools import setup, find_packages

setup(name="actonet",
    version="0.90",
    description = "Abduction-based control of Boolean networks fixpoints",
    install_requires = [
        "algorecell_types",
        "asprin",
        "colomoto_jupyter",
    ],
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords="computational systems biology",

    include_package_data = True,
    packages = find_packages()
)

